/* decin64.c : Routine decin64() converts input decimal digits to
   a 64-bit unsigned integer (as high and low 32-bit parts).
   This routine requires type "unsigned int" to be 32 bits (4 bytes).

   (c) Copyright 2005 by Dave Edwards.  All rights reserved.
   See the accompanying license.txt file.

   Usage:    int rc;
             unsigned int valhi,vallo;
             unsigned char *str;

             rc=decin64(str,&valhi,&vallo);
             if(rc<0) {...error...}

   Input is a null-terminated string.  Leading and trailing blanks
   are ignored.  Leading zeros do not count towards the limit of
   19 significant digits.  At least 1 digit must be present.
   Minus (-) and plus (+) signs are not allowed.

   The high-order 32 bits of the result are returned in the valhi arg.
   The low-order 32 bits of the result are returned in the vallo arg.

   Function return value rc:
     0   No errors.
     -1  Error: No digits. Input string is null or all blanks.
     -2  Error: More than 19 significant digits.
     -3  Error: Invalid character.
     -4  Error: Type "unsigned int" is not size 4.
   Note: If an error occurs, valhi and vallo are set to 0.

   Examples:  rc=decin64("5123456789",&valhi,&vallo);
              sets rc=0, valhi=0x00000001, vallo=0x3161BF15

              rc=decin64("  009123456789  ",&valhi,&vallo);
              sets rc=0, valhi=0x00000002, vallo=0x1FCCE715

   Tested in:  Windows (Win32)
               WatC on MUSIC/SP  (ccde:watc.decin*)
               Red Hat Linux 8.0 (gcc 3.2 compiler)

   See also:  c:\misc\subs\mysubs.h  - contains prototype
              c:\misc\sim390\dechexin.c  - decin(), 32-bit
              c:\misc\subs\readuint.c
              c:\misc\subs\readuilm.c
              c:\misc\testpgms\decin64t.c  - test program

Change Log:
27NOV2005 - Written by Dave Edwards.
01DEC2005 - Use faster method of multiplying by 10: shift and add.
   Is about 3 times faster than old method (9 adds). (DE)
*/

int decin64(unsigned char *str, unsigned int *pvalhi,
            unsigned int *pvallo)
{
unsigned int valhi,vallo,nsignifdigs,trailsw,digsw,carry,n1;
unsigned char ch;

*pvalhi=0; *pvallo=0;  /* set 0 result in case error */
if(sizeof(valhi)!=4) return -4;
valhi=0; vallo=0;  /* accumulate the 64-bit result here */
nsignifdigs=0;  /* # of digits, not counting leading 0's */
digsw=0;  /* set to 1 when any digit found */
trailsw=0;  /* set to 1 after a non-leading blank is found */
for(;;)
  { ch=*(str++);  /* get next char from the string */
    if(ch==0) break;  /* exit loop if end of string */
    if(ch==' ')
      { if(digsw) trailsw=1;  /* maybe it's a trailing blank */
        continue;  /* skip leading and trailing blanks */
      }
    if(trailsw) return -3;  /* error: digit(s) followed by blank(s)
                      followed by a nonblank */
    digsw=1;
    if(ch=='0' && nsignifdigs==0) continue;  /* skip leading 0's */
    /* check for a valid digit */
    if(ch<'0' || ch>'9') return -3;  /* bad digit */
    nsignifdigs++;
    if(nsignifdigs>19) return -2; /* result perhaps would not fit
                            in 64 bits ( 10**20 >= 2*64 ) */
    /* Accumulate the digit: */
    if(nsignifdigs<=9) vallo=vallo*10+(ch-'0');  /* the easy case */
    else
      { /* We must do 64-bit arithmetic.
             If A and B are 32-bit unsigned integers, A+B (as a possibly
             truncated 32-bit result) has a carry into the next higher
             bit position if and only if A+B (32-bit value) < B. */
        /* Multiply vallo by 10, by shift and add, keeping the carry.
           Use the formula:  10x = 2(4x+x)  (mutiply by shifting) */
        n1=vallo;
        carry=n1>>30;  /* multiply by 4: result to (carry,vallo) */
        vallo<<=2;
        vallo+=n1;  /* add n1 */
        if(vallo<n1) carry++;
        carry=(carry<<1)|(vallo>>31);  /* multiply by 2 */
        vallo<<=1;  /* result of vallo*10 is now in (carry,vallo) */
        n1=ch-'0';
        vallo+=n1;  /* add the digit to vallo, keeping the carry */
        if(vallo<n1) carry++;
        valhi=valhi*10+carry;  /* get final 64-bit result in (valhi,vallo) */
      }
  }
if(digsw==0) return -1;  /* no digits */
*pvalhi=valhi;  /* return the value */
*pvallo=vallo;
return 0;  /* success */
}  /* end of decin64 */
