/* rdline.c - rdline() function for Unix: reads next line from standard
   input (stdin) into a caller-supplied buffer, with ability to
   suppress echo of the input characters if stdin is the terminal.
   The line is returned as a null-terminated string in the buffer.
   There is no Newline (\n) character at the end of the string,
   but the Newline char at the end of the input line IS read.
   This function can be used to read a password or pass phrase,
   without echo.  The password or phrase can contains any special
   characters, including leading and trailing blanks.
   Note that if the buffer is too small, the line is truncated
   without warning, and the rest of the line is flushed (read and
   discarded).  If an error occurs, a null string (0 length) is
   put into the buffer.  See the warning below if scanf() is used
   before rdline().  The max length string that can be read is
   (size of buffer)-1.  If no echo (and stdin is a terminal), a
   Newline char ('\n') is written to stdout before returning; this
   can be suppressed by an option bit (see below).

   (c) Copyright 2005 by Dave Edwards.  All rights reserved.
   See the accompanying license.txt file.

   *** This function works only on a Unix/Linux system.
       It does not work on Win32 (use readpw() on Win32).
       Tested on:  Read Hat Linux 8.0 (gcc 3.2 compiler).

   Usage:   int rc,opt;
            unsigned char buf[n];   (n should be 2 or more. Max # of
                              text chars that can be returned is n-1.)

            opt=0 or 2 or 8;   (0,8: no echo. 2: normal echo)
            rc=rdline(buf,sizeof(buf),opt);
            if(rc<0) ...error...

   Option bits (additive) in opt arg:
            1   Not supported. Is ignored. (This bit is for
                  compatibility with readpw(), where it requests
                  echo of * for each char entered. In rdline(),
                  it results in no echo, unless bit 2 is on.)
            2   Normal echo of entered characters.
            4   Not supported. Is ignored. (This bit is for
                  compatibility with readpw(), where it displays buf
                  and allows editing of it.)
            8   When using no echo, suppress the final Newline '\n' char
                  that is otherwise normally put out after reading the
                  line. This is useful if the output of the calling
                  prog is being piped to another program, and you don't
                  want the Newline to be piped.  For example, the
                  calling prog may just be a prog that reads a password
                  (with no echo) and writes it to stdout for piping to
                  another program.  See the readpass program.
            Other bits: Ignored.
            Notes: - For opt, the same values as for Win32 readpw() can
                     be used (but option bits 1 and 4 are ignored).
                   - opt=0 or 8 results in no echo of the entered chars.
                     opt=0 is normally used to read a password or phrase.
                     This is implemented by temporarily setting the
                     terminal options to suppress echo of entered chars.

   Function return value rc:
     0   No error.
     -1  Buffer size is too small (0 or negative). buf is not changed.
           Note: bufsize=1 is not an error, but no read from stdin
           is done, and an empty string is returned in buf.
     -2  Unsupported option bit. A null (zero-length) string is
           returned in buf. (Currently this error does not occur.)
     -3  Some other error. A null string is returned in buf.
           (Currently this error does not occur.)

   Notes:
    - WARNING: scanf(), when used to read an integer, for example,
      leaves at least an unread Newline char from the current line.
      Therefore if rdline() is to be used after scanf(), flush the
      current line, up to and including the next '\n' char, by
      calls to fgetc(stdin).  See the test program.
    - If a password is read, opt=0 (no echo) should be used, and
      the buffer should be cleared as soon as the password is no
      longer needed.  The password may contain embedded blanks and
      other special characters.  buf should be large enough that
      the password is not truncated, i.e. at least (longest password)+1.
    - While the line is being entered, the normal editing keys can
      be used.  However, the Backspace key is the only editing key
      that is normally used.

   Standard library functions used:
       fileno, isatty, tcgetattr, tcsetattr, fgets, fgetc, putchar

   See also:  c:\misc\minteg\readpw.c  - similar function for Win32
              c:\misc\subs\rdnxtlin.c  - read next line from a file (Win32)
              c:\misc\linuxprogs\readpass.c  - Unix prog: reads a password

Change Log:
27nov2005 - Written by Dave Edwards
30nov2005 - If no echo, put out final Newline ('\n') before returning.
 - Added option bit 8. (DE)
*/

#include <stdio.h>
#include <string.h>
#include <termio.h>   /* for tcgetattr(), tcsetattr() */
#include <unistd.h>   /* for isatty() */

int rdline(unsigned char *buf, int bufsize, int opt)
{
int n1,goteol;
int noechosw;  /* nonzero if "no echo" option */
int termfd;  /* file descriptor of stdin */
struct termios old,new;  /* terminal settings */

if(bufsize<1) return -1;  /* buf is too small */
noechosw=2-(opt&2);
termfd=fileno(stdin);  /* get file descriptor of stdin stream */
if(!isatty(termfd)) noechosw=0;
if(noechosw)
  { /* Set terminal options to suppress echo of entered chars.
       Save the current settings so we can restore them later. */
    if(tcgetattr(termfd,&old)) noechosw=0;  /* error from tcgetattr() */
    else
      { new=old;
        new.c_lflag&=~ECHO;  /* turn off echo */
        if(tcsetattr(termfd,TCSAFLUSH,&new)) noechosw=0;  /* possible
                               error setting the options */
      }
  }
/* Read the line, max bufsize-1 chars. */
buf[0]=0;  /* (not really necessary) */
if(fgets(buf,bufsize,stdin)==NULL) buf[0]=0;  /* EOF and no chars read,
                        or error: set null string */
          /* Note: fgets() seems to return NULL if bufsize is 1. */
else
  { n1=strlen(buf);  /* get # of chars read (may include Newline) */
    goteol=0;
    if(n1>0) if(buf[n1-1]=='\n')
               { buf[n1-1]=0; goteol=1; }  /* remove Newline */
    if(!goteol)
      { /* If we did not get end-of-line, flush rest of the line.
           E.g. this can occur if buf is small and line is long. */
        for(;;)
          { n1=fgetc(stdin);  /* read next char */
            if((char)n1=='\n' || n1==EOF) break;
          }
      }
  }
/* Restore the terminal settings if we changed them. */
if(noechosw)
  { tcsetattr(termfd,TCSAFLUSH,&old);
  }
/* If no echo, put out final Newline char, since the user pressed
   the Return key. The option bit 8 suppresses this. */
if(noechosw && (opt&8)==0) putchar('\n');
return 0;
}  /* end of rdline */
