/* wipezero utility - Unix/Linux version:

   Writes a specified number of zero bytes to an existing file,
   possibly enlarging the file.  The starting byte displacement for
   the write can be specified.  This prog is intended for
   initialization purposes, to set the size of the file.  This prog
   does not delete the file.  The file must already exist.

   (c) Copyright 2005 by Dave Edwards.  All rights reserved.
   See the accompanying license.txt file.

   This program runs as a Unix or Linux command.
   Tested on:  Red Hat Linux 8.0 (gcc 3.2 C compiler).

   Usage:  wipezero filename

   Notes and restrictions:
    - This program works only with files up to 2 GB (2**31-1 bytes)
      in size, since the size and offset variables are signed 32-bit
      integers.  To support large files, the 64-bit functions
      open64(), fstat64(), and lseek64() would have to be used,
      along with the stat64 structure.
      However, the input and message functions used, decin64() and
      cnvsize(), are set up to handle 64-bit values.
    - The part of the file that is not written to is left unchanged.
      Therefore the program can be used to enlarge a file, without
      destroying the existing data.

   Other source files used:
      rdline.c
      cnvsize.c
      decin64.c

   To compile:  gcc -o wipezero wipezero.c rdline.c cnvsize.c decin64.c

Change Log:
28nov2005 - Written by Dave Edwards, based on the Win32 version.
*/

#include <stdio.h>
#include <string.h>
#include <fcntl.h>    /* for open() */
#include <unistd.h>   /* for write(), close(), lseek()
#include <sys/types.h>  /* for fstat() structure */
#include <sys/stat.h>  /* for fstat() */
#include <errno.h>

typedef unsigned int ULONG;
typedef unsigned char UCHAR;

/* Prototypes for external functions called */
int decin64(unsigned char *str, unsigned int *pvalhi,
               unsigned int *pvallo);
void rdline(unsigned char buf[], int buflen, int opt);
void cnvsize(ULONG sizehi, ULONG sizelo, UCHAR buf[40], ULONG sizbuf,
             ULONG opts);

int main(int numarg, char *cmdarg[])
{
ULONG n1,filesizelo,filesizehi,wrsize,left,wrlen,bufsize;
ULONG displlo,displhi,nwritten;
int rc;
ULONG dumhi,dispendsw;
UCHAR *fn;  /* file name (or path name) */
UCHAR ansbuf[25],sizebuf[40];
int hFile;  /* file handle (descriptor) */
struct stat statinfo;  /* file info */
UCHAR buf[32768];  /* buffer */

/* get the parameters from the command */
numarg--;  /* get actual no. of parameters */
if(numarg<1)   /* give help if no parms */
  { printf("wipezero writes a specified # of zero bytes to a file, at a\n");
    printf("specified byte displacement.  The original file size may be increased.\n");
    printf("Usage:  wipezero filename\n");
    return 1;
  }
fn=cmdarg[1];  /* file name */

printf("Will write zeros to file %s\n",fn);

/* Open the file for writing */
hFile=open(fn, O_WRONLY, 0);
if(hFile<0)
  { printf("** Error %d opening the file: %s\n",errno,strerror(errno));
    return 2;
  }

/* Get size of file */
rc=fstat(hFile,&statinfo);
if(rc)
  { printf("** Error %d getting file size: %s\n",errno,strerror(errno));
    return 5;
  }
n1=(ULONG)statinfo.st_size;
if(n1>=0x80000000)
  { printf("** Invalid file size. Max allowed is about 2 GB.\n");
    return 5;
  }
filesizelo=n1; filesizehi=0;
cnvsize(filesizehi,filesizelo,sizebuf,sizeof(sizebuf),1);
printf("Original file size: %s\n",sizebuf);

/* Ask for displacement for write */
printf("Enter byte displacement for zeroing (max 2**31-1),\n"
       "or 0 (start of file), or e (end of file): ");
dispendsw=0;
rdline(ansbuf,sizeof(ansbuf),2);
if((ansbuf[0]=='e' || ansbuf[0]=='E') && ansbuf[1]==0)
  { displlo=filesizelo;
    displhi=filesizehi;
    dispendsw=1;
  }
else
  { rc=decin64(ansbuf,&displhi,&displlo);
    if(rc || displhi!=0 || displlo>=0x80000000)
      { printf("** Invalid value. Max is 2**31-1 = 2147483647\n");
        return 7;
      }
  }
if(dispendsw) rc=lseek(hFile,0,SEEK_END);
else rc=lseek(hFile,(long int)displlo,SEEK_SET);
if(rc==-1)
  { printf("** Error %d setting file pointer: %s\n",errno,strerror(errno));
    return 8;
  }
cnvsize(displhi,displlo,sizebuf,sizeof(sizebuf),1);
printf("Using displacement: %s\n",sizebuf);

/* Ask for length to be written */
printf("Enter number of zero bytes to be written: ");
rdline(ansbuf,sizeof(ansbuf),2);
rc=decin64(ansbuf,&dumhi,&wrsize);
if(rc || dumhi!=0 || wrsize>=0x80000000)
  { printf("** Invalid value. Max is 2**31-1 = 2147483647\n");
    return 9;
  }
printf("Using write length %lu\n",wrsize);

printf("Enter y to start zeroing: ");
rdline(ansbuf,sizeof(ansbuf),2);
if(ansbuf[0]!='y' && ansbuf[0]!='Y')
  { printf("** Cancelled, file not changed\n");
    return 10;
  }

bufsize=sizeof(buf);
memset(buf,0,bufsize);  /* init buffer to zeros */

for(left=wrsize;left>0;left-=wrlen)  /* write loop */
  { wrlen=left; if(wrlen>bufsize) wrlen=bufsize;  /* # of bytes to write */
    rc=write(hFile,buf,(int)wrlen);
    if(rc<0)
      { printf("** Error %d writing to the file: %s\n",
              errno,strerror(errno));
        close(hFile);
        return 6;
      }
    nwritten=(ULONG)rc;
    if(nwritten!=wrlen)
      { printf("** Error: write() requested %lu bytes to be written, "
               "but only %lu bytes were actually written\n",
               wrlen,nwritten);
        close(hFile);
        return 6;
      }
  }

if(close(hFile))
  { printf("** Error %d closing the file: %s\n",errno,strerror(errno));
    return 4;
  }
printf("File area zeroed\n");
return 0;  /* normal end */
}  /* end of main */
